/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CAbstractBodyContext.java,v 1.2 2002/03/08 21:54:12 mrobilla Exp $
 */

package at.dms.kjc;

import java.util.Hashtable;

import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents a local context during checkBody
 * It follows the control flow and maintain informations about
 * variable (initialized, used, allocated), exceptions (thrown, catched)
 * It also verify that context is still reachable
 *
 * There is a set of utilities method to access fields, methods and class
 * with the name by clamping the parsing tree
 * @see CContext
 * @see CCompilationUnitContext
 * @see CClassContext
 * @see CMethodContext
 * @see CConstructorContext
 * @see CInitializerContext
 * @see CBodyContext
 * @see CBlockContext
 */
public class CAbstractBodyContext extends CContext {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a block context, it supports local variable allocation
   * throw statement and return statement
   * This starts a new context in a method
   * @param	parent		the parent context, it must be different
   *				than null except if called by the top level
   */
  protected CAbstractBodyContext(CAbstractBodyContext parent) {
    super(parent);

    flowState = parent.flowState;
    throwables = parent.throwables;
  }

  /**
   * Construct a block context, it supports local variable allocation
   * throw statement and return statement
   * This starts a new context in a method
   * @param	parent		the parent context, it must be different
   *				than null except if called by the top level
   */
  protected CAbstractBodyContext(CMethodContext parent) {
    super(parent);

    //flowState = 0;
    throwables = parent.getThrowables();
    assert(this instanceof CBlockContext);
  }

  /**
   * Verify everything is okay at the end of this context
   */
  public void close(TokenReference ref) {
    informParent();
  }

  /**
   * Verify everything is okay at the end of this context
   */
  public void informParent() {
    // inform parent
    if (parent instanceof CAbstractBodyContext) {
      CAbstractBodyContext	parent = (CAbstractBodyContext)this.parent;
      
      if (variableInfo != null) {
	int	parentPosition = parent.getBlockContext().localsIndex();
	
	for (int i = 0; i < parentPosition; i++) {
	  int	info = variableInfo.getInfo(i);
	  
	  if (info != 0) {
	    parent.setVariableInfo(i, info);
	  }
	}
      }
      
      parent.flowState = flowState;
    }

    if (fieldInfo != null) {
      int	parentPosition = parent.getClassContext().getCClass().getFieldCount();
      
      for (int i = 0; i < parentPosition; i++) {
	int	info = fieldInfo.getInfo(i);
	
	if (info != 0) {
	  parent.setFieldInfo(i, info);
	}
      }
    }
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * isLeftSide
   * @return	true if this context is on a left side of an assignement
   */
  public boolean isInLoop() {
    return (flowState & FLO_INLOOP) != 0;
  }

  /**
   *
   */
  public void setReachable(boolean reachable) {
    if (reachable) {
      flowState &= ~FLO_UNREACHABLE;
    } else {
      flowState |= FLO_UNREACHABLE;
    }
  }

  /**
   *
   */
  public void addReturn() {
    setReachable(false);
  }

  /**
   *
   */
  public void addBreak() {
    flowState |= FLO_UNREACHABLE | FLO_BREAKED;
  }

  /**
   *
   */
  public void addBreak(CBodyContext context) {
    addBreak();
  }

  /**
   *
   */
  public void addContinue() {
    flowState |= FLO_UNREACHABLE | FLO_CONTINUED;
  }

  /**
   * isReachable
   * @return if this context is still reachable
   */
  public boolean isReachable() {
    return (flowState & FLO_UNREACHABLE) == 0;
  }

  /**
   * @return	if this context was set unreachable by a break stmt
   */
  public boolean isBreaked() {
    return (flowState & FLO_BREAKED) != 0;
  }

  /**
   * @return	if this context was set unreachable by a continue stmt
   */
  public boolean isContinued() {
    return (flowState & FLO_CONTINUED) != 0;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (VARIABLE)
  // ----------------------------------------------------------------------

  public int localsPosition() {
    return ((CAbstractBodyContext)parent).localsPosition();
  }

  /**
   * Returns the statement with the specified label.
   */
  public JLabeledStatement getLabeledStatement(String label) {
    return parent instanceof CAbstractBodyContext ? // $$$ remove ths test 
      ((CAbstractBodyContext)parent).getLabeledStatement(label) :
      null;
  }

  /**
   * Returns the nearest breakable statement.
   */
  public JStatement getNearestBreakableStatement() {
    return parent instanceof CAbstractBodyContext ? // $$$ remove ths test 
      ((CAbstractBodyContext)parent).getNearestBreakableStatement() :
      null;
  }

  /**
   * Returns the nearest continuable statement.
   */
  public JStatement getNearestContinuableStatement() {
    return parent instanceof CAbstractBodyContext ?
      ((CAbstractBodyContext)parent).getNearestContinuableStatement() :
      null;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (INFOS)
  // ----------------------------------------------------------------------

  /**
   * @param	index		The variable position in method array of local vars
   * @param	info		The information to add
   *
   * We make it a local copy of this information and at the end of this context
   * we will transfer it to the parent context according to controlFlow
   */
  public void setVariableInfo(int index, int info) {
    if (variableInfo == null) {
      variableInfo = new CVariableInfo();
    }
    if (variableInfo.getInfo(index) != 0) {
      variableInfo.setInfo(index, info);
    } else {
      variableInfo.setInfo(index, info | getVariableInfo(index));
    }
  }

  /**
   * @param	var		the definition of a variable
   * @return	all informations we have about this variable
   */
  public int getVariableInfo(int index) {
    if (variableInfo == null) {
      return ((CAbstractBodyContext)parent).getVariableInfo(index);
    }
    int		info = variableInfo.getInfo(index);

    if (info == 0) {
      info = ((CAbstractBodyContext)parent).getVariableInfo(index);
    } 

    return info;
  }

  /**
   * @param	index		The field position in method array of local vars
   * @param	info		The information to add
   *
   * We make it a local copy of this information and at the end of this context
   * we will transfert it to the parent context according to controlFlow
   */
  public void setFieldInfo(int index, int info) {
    if (fieldInfo == null) {
      fieldInfo = new CVariableInfo();
    }
    if (fieldInfo.getInfo(index) != 0) {
      fieldInfo.setInfo(index, info);
    } else {
      fieldInfo.setInfo(index, info | getFieldInfo(index));
    }
  }

  /**
   * @param	var		the definition of a field
   * @return	all informations we have about this field
   */
  public int getFieldInfo(int index) {
    if (fieldInfo == null) {
      return parent.getFieldInfo(index);
    }
    int		info = fieldInfo.getInfo(index);

    if (info == 0) {
      info = parent.getFieldInfo(index);
    } 

    return info;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (LOOKUP)
  // ----------------------------------------------------------------------

  /**
   * lookupClass
   * search for a class with the provided type parameters
   * @param	name		method name
   * @return	the class if found, null otherwise
   * @exception UnpositionedError	this error will be positioned soon
   *
  public CClassType lookupClass(String name) throws UnpositionedError {
    return getParentContext().lookupClass(name);
  }

  /**
   * lookupField
   * @param	ident		the name of the field
   * @return	a variable from an ident in current context
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CField lookupField(String ident) throws UnpositionedError {
    return getClassContext().lookupField(ident);
  }

  /**
   * lookupLocalVariable
   * @param	ident		the name of the variable
   * @return	a variable from an ident in current context
   */
  public JLocalVariable lookupLocalVariable(String ident) {
    return parent.lookupLocalVariable(ident);
  }

  /**
   * lookupOuterLocalVariable
   * @param	ident		the name of the variable
   * @return	a variable from an ident in upperclass context
   */
  public JExpression lookupOuterLocalVariable(TokenReference ref, String ident) {
    CClassContext	owner = getClassContext();

    return owner.lookupOuterLocalVariable(ref, ident);
  }

  // ----------------------------------------------------------------------
  // THROWABLES
  // ----------------------------------------------------------------------

  /**
   * @param	throwable	the type of the new throwable
   */
  public void addThrowable(CThrowableInfo throwable) {
    if (throwables == null) {
      throwables = new Hashtable();      
    }
    throwables.put(throwable, throwable);
  }

  /**
   * @return the list of exception that may be thrown
   */
  public Hashtable getThrowables() {
    return throwables == null ? EMPTY : throwables;
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private static final Hashtable	EMPTY = new Hashtable(1);

  // Flow infos // hard copy
  protected	int		flowState;
  protected	CVariableInfo	variableInfo;
  protected	CVariableInfo	fieldInfo;
  protected	Hashtable	throwables;

  public static final int	FLO_UNREACHABLE = 1 << 0;
  public static final int	FLO_BREAKED = 1 << 1;
  public static final int	FLO_CONTINUED = 1 << 2;

  public static final int	FLO_INLOOP = 1 << 3;
}
