/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClass.java,v 1.2 2002/03/08 21:54:13 mrobilla Exp $
 *
 * Modified by Martin P. Robillard for Jex 26/06/2001
 */

package at.dms.kjc;

import java.io.File;
import java.io.IOException;

import java.util.Hashtable;
import java.util.Vector;

import at.dms.classfile.*;
import at.dms.compiler.Compiler;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.Utils;
import at.dms.util.VectorCache;

/**
 * This class represents the exported members of a class (inner classes, methods and fields)
 */
public abstract class CClass extends CMember {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a class export from file
   */
  public CClass(Compiler compiler,
		CClass owner,
		String sourceFile,
		int modifiers,
		String ident,
		String qualifiedName,
		CClassType superClass,
		boolean deprecated)
  {
    super(owner, modifiers, ident, deprecated);

    this.compiler = compiler;
    this.sourceFile = sourceFile;
    this.qualifiedName = qualifiedName;
    this.superClass = superClass;

    int		cursor = qualifiedName.lastIndexOf('/');

    this.packageName = cursor > 0 ? qualifiedName.substring(0, cursor).intern() : "";
  }

  /**
   * Ends the definition of this class
   */
  public void close(CClassType[] interfaces,
		    CField[] fields_A,
		    Hashtable fields_H,
		    CMethod[] methods_A,
		    Hashtable methods_H)
  {
    this.interfaces = interfaces;
    this.fields = new CFieldTable(fields_A, fields_H);
    this.methods_A = methods_A;
    this.methods_H = methods_H;
    assert(interfaces != null);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------
  /**
   * Returns the compiler 
   */
  public Compiler getCompiler() {
    return compiler;
  }

  /**
   * @return	the interface
   */
  public CClass getCClass() {
    return this;
  }

  /**
   * @return true if this class is an inner class
   */
  public boolean isInnerClass() {
    return getOwner() != null;
  }

  /**
   * @return	true if this member is abstract
   */
  public boolean isAbstract() {
    return hasFlag(getModifiers(), ACC_ABSTRACT);
  }

  /**
   * @return true if this class is anonymous
   */
  public boolean isAnonymous() {
    return getIdent().length() == 0;
  }

  /**
   * @return true if this class is has an outer this
   */
  public boolean hasOuterThis() {
    return hasOuterThis;
  }

  /**
   * Sets hasOuterThis
   */
  public void setHasOuterThis(boolean hasOuterThis) {
    this.hasOuterThis = hasOuterThis;
  }

  /**
   * @return true if this class is an interface
   */
  public boolean isInterface() {
    return hasFlag(getModifiers(), ACC_INTERFACE);
  }

  /**
   * @return the full name of this class
   */
  public String getQualifiedName() {
    return qualifiedName;
  }

  /**
   * @return the name of the package of the package this class belongs to
   */
 public String getPackage() {
    return packageName;
  }

  /**
   * Returns the super class of this class
   */
  public CClass getSuperClass() {
    return superClass == null ? null : superClass.getCClass();
  }

  /**
   * Sets the super class of this class
   */
  public void setSuperClass(CClassType superClass) {
    this.superClass = superClass;
  }

  /**
   * Returns the type of this class
   */
  public CClassType getType() {
    if (type == null) {
      type = CClassType.lookup(getQualifiedName());
    }
    return type;
  }

  /**
   * @return the source file of this class
   */
  public String getSourceFile() {
    return sourceFile;
  }

  /**
   * @param	ident		the name of the field
   * @return	the field
   */
  public CField getField(String ident) {
    return fields.lookup(ident);
  }

  public int getFieldCount() {
    return fields.size();
  }

  /**
   * @return the InnerClasses
   */
  public CClassType[] getInnerClasses() {
    return innerClasses;
  }

  /**
   * End of first pass, we need innerclassesinterfaces
   */
  public void setInnerClasses(CClassType[] inners) {
    this.innerClasses = inners;
  }

  /**
   * descendsFrom
   * @param	from	an other CClass
   * @return	true if this class inherit from "from" or equals "from"
   */
  public boolean descendsFrom(CClass from) {
    if (from == this) {
      return true;
    } else if (superClass == null) {
      // java/lang/object
      return false;
    } else if (getSuperClass().descendsFrom(from)) {
      return true;
    } else {
      for (int i = 0; i < interfaces.length; i++) {
	if (interfaces[i].getCClass().descendsFrom(from)) {
	  return true;
	}
      }

      return false;
    }
  }

  /**
   * Returns true iff this class is defined inside the specified class
   * @param	outer		another class
   */
  public boolean isDefinedInside(CClass outer) {
    if (this == outer) {
      return true;
    } else if (getOwner() == null) {
      return false;
    } else {
      return getOwner().isDefinedInside(outer);
    }
  }

  // ----------------------------------------------------------------------
  // LOOKUP
  // ----------------------------------------------------------------------

  /**
   * This can be used to see if a given class name is visible
   * inside of this file.  This includes globally-qualified class names that
   * include their package and simple names that are visible thanks to some
   * previous import statement or visible because they are in this file.
   * If one is found, that entry is returned, otherwise null is returned.
   * @param	name		a TypeName (6.5.2)
   */
  public CClass lookupClass(String name) {
    if (innerClasses != null && (name.indexOf('/') == -1)) {
      for (int i = 0; i < innerClasses.length; i++) {
	if (innerClasses[i].getCClass().getIdent().equals(name)) {
	  return innerClasses[i].getCClass();
	}
      }
    }
    if (superClass != null) {
      return superClass.getCClass().lookupClass(name);
    }

    return null;
  }

  /**
   * Searches a field in current class and parent hierarchy as needed
   * @param	name		the simple name of the field
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CField lookupField(String name) throws UnpositionedError {
    CField	field = getField(name);

    if (field != null) {
      return field;
    } else {
      return lookupSuperField(name);
    }
  }

  /**
   * Searches a field in current class and parent hierarchy as needed
   * @param	name		the simple name of the field
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CField lookupSuperField(String name) throws UnpositionedError {
    CField	field = getField(name);

    if (superClass != null) {
      // java.lang.object / interfaces
      field = getSuperClass().lookupField(name);
    }

    for (int i = 0; i < interfaces.length; i++) {
      CField	newField = interfaces[i].getCClass().lookupField(name);

      if ((field != null) && (newField != null) && (field.getOwner() != newField.getOwner())) {
	throw new UnpositionedError(KjcMessages.FIELD_AMBIGUOUS, name);
      }

      if (newField != null) {
	field = newField;
      }
    }

    return field;
  }

  /**
   * lookupMethod
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CMethod lookupMethod(String name, CType[] params) throws UnpositionedError {
    CMethod	method = getMatchingMethod(name, params);

    if (method != null) {
      return method;
    } else {
      return lookupSuperMethod(name, params);
    }
  }

  /**
   * lookupSuperMethod
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CMethod lookupSuperMethod(String name, CType[] params) throws UnpositionedError {
    CMethod	method;

    if (superClass == null) {
      method = null;
    } else {
      method = getSuperClass().lookupMethod(name, params);
    }

    for (int i = 0; i < interfaces.length && method == null; i++) {
      method = interfaces[i].getCClass().lookupMethod(name, params);
    }

    return method;
  }

  /**
   * getMatchingMethod
   * search for a matching method with the provided type parameters
   * Linear search (if method exists)
   * @param	ident		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  private CMethod getMatchingMethod(String ident, CType[] params) throws UnpositionedError {
    CMethod	method = null;
    boolean	lastMethodAmbiguous = false;

    if (methods_H != null &&
	ident != JAV_CONSTRUCTOR &&
	ident != JAV_INIT &&
	ident != JAV_STATIC_INIT &&
	!methods_H.containsKey(ident)) {
      return null;
    }

    for (int i = 0; i < methods_A.length; i++) {
      CMethod	current = methods_A[i];

      if (current.isMatching(ident, params)) {
	if (method == null || current.isMoreSpecificThan(method)) {
	  method = current;
	  lastMethodAmbiguous = false;
	} else if (method.isMoreSpecificThan(current)) {
	  lastMethodAmbiguous = false;
	} else {
	  lastMethodAmbiguous = true;
	}
      }
    }

    if (lastMethodAmbiguous) {
      throw new UnpositionedError(KjcMessages.METHOD_AMBIGOUS, ident);
    }

    return method;
  }

  /**
   * Returns a list of abstract methods
   */
  public CMethod[] getAbstractMethods() {
    Vector v = VectorCache.request(VectorCache.SMALL);
    collectAbstractMethods(v);
    CMethod[] methods = new CMethod[v.size()];
    v.copyInto(methods);
    VectorCache.release(v);
    return methods;
  }

  /**
   * Returns a list of interface methods
   */
  public CMethod[] getInterfaceMethods() {
    Vector v = VectorCache.request(VectorCache.SMALL);
    collectInterfaceMethods(v);
    CMethod[] methods = new CMethod[v.size()];
    v.copyInto(methods);
    VectorCache.release(v);
    return methods;
  }

  /**
   * collectAbstractMethods
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   */
  public void collectAbstractMethods(Vector v) {
    for (int i = 0; i < methods_A.length; i++) {
      CMethod	current = methods_A[i];

      if (current.isAbstract()) {
	v.addElement(current);
      }
    }

    if (superClass != null) {
      // java.lang.object
      getSuperClass().collectAbstractMethods(v);
    }
  }

  /**
   * collectInterfaceMethods
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   */
  public void collectInterfaceMethods(Vector v) {
    if (isInterface()) {
      for (int i = 0; i < methods_A.length; i++) {
	v.addElement(methods_A[i]);
      }
    }

    for (int i = 0; i < interfaces.length; i++) {
      CClass		parent;
      parent = interfaces[i].getCClass();
      assert(parent != null);
      parent.collectInterfaceMethods(v);
    }

    if (superClass != null) {
      // java.lang.object
      getSuperClass().collectInterfaceMethods(v);
    }
  }

  /**
   * @return the short name of this class
   */
  protected static String getIdent(String qualifiedName) {
    assert(qualifiedName != null);
    String	syntheticName;

    int		cursor = qualifiedName.lastIndexOf('/');
    if (cursor > 0) {
      syntheticName = qualifiedName.substring(cursor + 1);
    } else {
      syntheticName = qualifiedName;
    }

    cursor = syntheticName.lastIndexOf('$');
    if (cursor > 0) {
      return syntheticName.substring(cursor + 1);
    } else {
      return syntheticName;
    }
  }

  /**
   * @return	true if this member is accessible
   */
  public boolean isAccessible(CClass from) {
    if ((from.isDefinedInside(this)) || isPublic() || isDefinedInside(from)) {
      return true;
    } else if (from.isInnerClass() && isAccessible(from.getOwner())) {
      return true;
    } else if (!isPrivate() && getPackage() == from.getPackage()) {
      return true;
    } else if (isProtected() && from.descendsFrom(this)) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * Add inner fields
   */
  public void addField(CSourceField field) {
    fields.add(field);
  }

  // ----------------------------------------------------------------------
  // GENERATE CLASSFILE INFO
  // ----------------------------------------------------------------------

  /**
   * Generate the code in a class file
   * @param	destination	the root directory of the class hierarchy
   */
    public void genClassFile(String destination) throws IOException, ClassFileFormatException
    {
	ClassConstant[] interfaces = new ClassConstant[this.interfaces.length];

	for (int i = 0; i < this.interfaces.length; i++)
	{
	    interfaces[i] = new ClassConstant(this.interfaces[i].getQualifiedName());
	}

	int          count = 0;
	MethodInfo[] pmethods = new MethodInfo[methods_A.length];

	for (int i = 0; i < methods_A.length; i++)
	{
	    pmethods[count++] = methods_A[i].genMethodInfo();
	}

	FieldInfo[] pfields = fields.buildFieldInfo();

	InnerClassInfo[] inners = new InnerClassInfo[innerClasses.length];
	for (int i = 0; i < inners.length; i++)
        {
	    inners[i] = new InnerClassInfo(innerClasses[i].getCClass().getQualifiedName(),
					   getQualifiedName(),
					   innerClasses[i].getCClass().getIdent(),
					   (short)innerClasses[i].getCClass().getModifiers());
	}

	String[]	classPath = Utils.splitQualifiedName(sourceFile, File.separatorChar);
	ClassInfo	classInfo = new ClassInfo((short)(getModifiers() & (~ACC_STATIC)),
						  getQualifiedName(),
						  superClass == null ? null : superClass.getCClass().getQualifiedName(),
						  interfaces,
						  pfields,
						  pmethods,
						  inners,
						  classPath[1],
						  false);
	classInfo.write(destination != null ? destination : classPath[0]);
    }

  // ----------------------------------------------------------------------
  // INNER CLASS SUPPORT
  // ----------------------------------------------------------------------

  /**
   * add synthetic parameters to method def
   */
  public CType[] genConstructorArray(CType[] params) {
    throw new RuntimeException();
  }

  /**
   * add synthetic parameters to method call
   */
  public void genSyntheticParams(CodeSequence code) {
    if (isInnerClass() && hasOuterThis()) {
      code.plantLoadThis();
    }
  }

  /**
   * add synthetic parameters to method call
   */
  public void genSyntheticParamsFromExplicitSuper(JExpression[] params, CodeSequence code) {
    if (isInnerClass() && hasOuterThis()) {
      JGeneratedLocalVariable var = new JGeneratedLocalVariable(null, 0, getOwner().getType(), JAV_OUTER_THIS, null);
      var.setPosition(params.length + 1);
      new JLocalVariableExpression(TokenReference.NO_REF, var).genCode(code);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE INITIALIZATIONS
  // ----------------------------------------------------------------------

  protected static CClass	CLS_UNDEFINED = new CSourceClass(null, null, TokenReference.NO_REF, 0, "<gen>", "<gen>", false);

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private final Compiler	compiler;
  private final String		sourceFile;
  private final	String		qualifiedName;
  private final	String		packageName;

    public  CClassType[]		interfaces; // MR Made public
  private CClassType		type;
  private CClassType[]		innerClasses;
  private CClassType		superClass;

  private boolean		hasOuterThis;

  private CFieldTable		fields;

  private CMethod[]		methods_A;
  // May be null if there is less than a certain amount of members
  private Hashtable		methods_H;
}
