/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JArrayInitializer.java,v 1.2 2002/03/08 21:54:23 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class implements a constant list of expressions used in array initialisation
 * The method isAssignableTo is redefined in order to follow (JLS 5.2)
 */
public class JArrayInitializer extends JExpression {

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param where the line of this node in the source code
   * @param v a vector of expression {e1, e2, e3, e4}
   */
  public JArrayInitializer(TokenReference where, JExpression[] elems) {
    super(where);

    this.elems = elems;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param	type		the type of this array
   *				assume type was checked
   * @exception	PositionedError		if the check fails
   */
  public void setType(CType type) {
    assert(type.checked());
    this.type = type;
  }

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    assert(type != null);
    assert(type.isArrayType());

    CType	elemsType = ((CArrayType)type).getIndexedType();

    if (elemsType.isArrayType()) {
      for (int i = 0; i < elems.length; i++) {
	if (elems[i] instanceof JArrayInitializer) {
	  ((JArrayInitializer)elems[i]).setType(elemsType);
	}
	elems[i] = elems[i].checkExpression(context);
	check(context, elems[i].isAssignableTo(elemsType),
	      KjcMessages.ARRAY_INIT_BADTYPE, elemsType, elems[i].getType());
      }
    } else {
      for (int i = 0; i < elems.length; i++) {
	elems[i] = elems[i].checkExpression(context);
	check(context, elems[i].isAssignableTo(elemsType),
	      KjcMessages.ARRAY_INIT_BADTYPE, elemsType, elems[i].getType());
	elems[i] = elems[i].convertType(elemsType, context);
      }
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitArrayInitializer(this, elems);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    JArrayDimsAndInits.allocArray(code,
				  (CArrayType)getType(),
				  new JExpression[] { new JOrdinalLiteral(getTokenReference(),
									  elems.length,
									  CStdType.Integer)});
    int		opcode;

    switch (((CArrayType)getType()).getIndexedType().getTypeID()) {
    case TID_CLASS:
    case TID_ARRAY:
      opcode = opc_aastore;
      break;
    case TID_INT:
      opcode = opc_iastore;
      break;
    case TID_LONG:
      opcode = opc_lastore;
      break;
    case TID_FLOAT:
      opcode = opc_fastore;
      break;
    case TID_DOUBLE:
      opcode = opc_dastore;
      break;
    case TID_BYTE:
    case TID_BOOLEAN:
      opcode = opc_bastore;
      break;
    case TID_CHAR:
      opcode = opc_castore;
      break;
    case TID_SHORT:
      opcode = opc_sastore;
      break;
    default:
      fail();
      opcode = -1;
    }

    for (int i = 0; i < elems.length; i++) {
      //if (i != elems.length - 1)
      code.plantInstruction(new NoArgInstruction(opc_dup));
      code.plantInstruction(new PushLiteralInstruction(i));
      elems[i].genCode(code);
      code.plantInstruction(new NoArgInstruction(opcode));
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CType		type;
  private	JExpression[]	elems;
}
