/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCharLiteral.java,v 1.2 2002/03/08 21:54:25 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.NumberParser;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * A simple character constant
 */
public class JCharLiteral extends JOrdinalLiteral {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   */
  public JCharLiteral(TokenReference where, String image) {
    super(where, image);
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   */
  public JCharLiteral(TokenReference where, char val) {
    super(where, "");

    value = new Character(val);
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   */
  public JCharLiteral(TokenReference where, Object val) {
    super(where, "");
    if (val instanceof Character) {
      value = (Character)val;
    } else {
      value = new Character((char)((Number)val).intValue());
    }
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   * @param	quoted		there is quote around image
   */
  public JCharLiteral(TokenReference where, String image, boolean quoted) {
    super(where, image.substring(1, image.length() - 1));
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return CStdType.Char;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Number numberValue() {
    return new Integer(value.charValue());
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Object getValue() {
    return value;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    if (image == null) {
      return this;
    }
    if (image.startsWith("\\u")) {
      value = new Character((char)NumberParser.decodeHexInt(false, image.substring(2)));
    } else if (image.startsWith("\\") && image.length() > 1) {
      if ((image.charAt(1) >= '0') && (image.charAt(1) <= '9')) {
	value = new Character((char)NumberParser.decodeOctInt(false, image.substring(1)));
	check(context, value.charValue() <= 0377,
	      KjcMessages.INVALID_OCTAL_CHAR, image);
      } else if (image.equals("\\b")) {
	value = new Character('\b');
      } else if (image.equals("\\r")) {
	value = new Character('\r');
      } else if (image.equals("\\t")) {
	value = new Character('\t');
      } else if (image.equals("\\n")) {
	value = new Character('\n');
      } else if (image.equals("\\f")) {
	value = new Character('\f');
      } else if (image.equals("\\\"")) {
	value = new Character('\"');
      } else if (image.equals("\\\'")) {
	value = new Character('\'');
      } else if (image.equals("\\\\")) {
	value = new Character('\\');
      } else {
	check(context, false, KjcMessages.INVALID_ESCAPE_SEQUENCE, image);
      }
    } else {
      value = new Character(image.charAt(0));
    }

    image = null;
    return this;
  }

  /**
   * convertType
   * changes the type of this expression to an other
   * @param  dest the destination type
   */
    public JExpression convertType(CType dest, CExpressionContext context) throws PositionedError
    {
	if (dest == CStdType.Char)
	{
	    return this;
	}
	else if (dest == CStdType.Byte)
        {
	    return new JOrdinalLiteral(getTokenReference(), new Byte((byte)value.charValue()), CStdType.Byte);
	}
	else if (dest == CStdType.Short)
	{
	    return new JOrdinalLiteral(getTokenReference(), new Short((short)value.charValue()), CStdType.Short);
	}
	else if (dest == CStdType.Integer)
	{
	    return new JOrdinalLiteral(getTokenReference(), new Integer((int)value.charValue()), CStdType.Integer);
	}
	else if (dest == CStdType.Long)
	{
	    return new JOrdinalLiteral(getTokenReference(), new Long((long)value.charValue()), CStdType.Long);
	}
	else if (dest == CStdType.Float)
        {
	    return new JRealLiteral(getTokenReference(), new Float((float)value.charValue()), CStdType.Float);
	}
	else if (dest == CStdType.Double)
	{
	    return new JRealLiteral(getTokenReference(), new Double((double)value.charValue()), CStdType.Double);
	}
	else
	{
	    System.out.println( getTokenReference() );
	    System.out.flush();
	    fail();
	    return null;
	}
    }

  /**
   * Can this expression be converted to the specified type by assignment conversion (JLS 5.2) ?
   * @param	dest		the destination type
   * @return	true iff the casting conversion is valid
   */
  public boolean isAssignableTo(CType dest) {
    return CStdType.Char.isAssignableTo(dest);
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    if (value != null) {
      p.visitCharLiteral(value.charValue());
    } else {
      p.visitCharLiteral(image.charAt(0)); // !!! false
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	Character		value;
}
