/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassDeclaration.java,v 1.2 2002/03/08 21:54:26 mrobilla Exp $
 */

package at.dms.kjc;

import java.util.Vector;

import at.dms.compiler.CWarning;
import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.VectorCache;

/**
 * This class represents a java class in the syntax tree
 */
public class JClassDeclaration extends JTypeDeclaration {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	context		the context in which this class is defined
   * @param	modifiers	the list of modifiers of this class
   * @param	ident		the short name of this class
   * @param	superName	the name of super class of this class
   * @param	interfaces	the name of this class's interfaces
   * @param	javadoc	is this class deprecated
   */
  public JClassDeclaration(TokenReference where,
			   int modifiers,
			   String ident,
			   String superName,
			   CClassType[] interfaces,
			   JFieldDeclaration[] fields,
			   JMethodDeclaration[] methods,
			   JTypeDeclaration[] inners,
			   JPhylum[] initializers,
			   JavadocComment javadoc,
			   JavaStyleComment[] comment) {
    super(where, modifiers, ident, interfaces, fields, methods, inners, initializers, javadoc, comment);
    this.superName = superName;
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Sets the super class
   */
  public void setSuperClass(String superName) {
    this.superName = superName;
  }

  /**
   * Sets the super class
   */
  public void setInterfaces(CClassType[] interfaces) {
    this.interfaces = interfaces;
  }

    /**
     * Second pass (quick), check interface looks good
     * Exceptions are not allowed here, this pass is just a tuning
     * pass in order to create informations about exported elements
     * such as Classes, Interfaces, Methods, Constructors and Fields
     * @return true iff sub tree is correct enought to check code
     * @exception	PositionedError	an error with reference to the source file
     */
    public void checkInterface(final CContext context) throws PositionedError
    {
	CClass	   superClass = null;
	CClassType type = null;
 
	if( superName == null )
	{
	    if (sourceClass.getType().equals(CStdType.Object))
	    {
		// java/lang/Object
		// superClass = null;
		// type = null;
	    }
	    else
	    {
		superClass = CStdType.Object.getCClass();
		type = CStdType.Object;
	    }
	}
	else
	{
	    // find name
	    try
	    {
		type = CClassType.lookup(superName);
		type.checkType(context);
		superClass = type.getCClass();
	    }
	    catch (UnpositionedError cue)
	    {
		// not yet defined, throw cue.addPosition(getTokenReference());
		type = new CClassNameType(superName)
		{
		    public void checkType(CContext ctxt) throws UnpositionedError
		    {
			CClassType temp = CClassType.lookup(superName);

			temp.checkType(context);
			setClass(temp.getCClass());
			qualifiedName = null;
		    }
		    };
	    }
	}

	// check access
	if (superClass != null)
	{
	    check(context, superClass.isPublic() || sourceClass.getPackage() == superClass.getPackage(),
		  KjcMessages.CLASS_ACCESSPARENT, superName);
	    check(context, !superClass.isFinal(), KjcMessages.CLASS_PARENT_FINAL, superName);
	    check(context, !superClass.isInterface(), KjcMessages.CLASS_EXTENDS_INTERFACE, superName);
	}

	CClassContext self = createContext(context);

	statInit = constructInitializers(true);
	if (statInit != null)
	{
	    statInit.checkInterface(self);
	}

	int	i = 0;
	for (; i < methods.length; i++)
	{
	    if (methods[i] instanceof JConstructorDeclaration)
	    {
		// usually constructors are at the top of the file
		// so its not as inefficient as it's look
		break;
	    }
	}
	if (i == methods.length && getDefaultConstructor() == null)
	{
	    setDefaultConstructor(constructDefaultConstructor());
	}
	
	instanceInit = constructInitializers(false);
	if (instanceInit != null)
	{
	    instanceInit.checkInterface(self);
	}
	
	getCClass().setSuperClass(type);
	super.checkInterface(context, type);
    }

  /**
   * Check that initializers are correct
   * @exception	PositionedError	an error with reference to the source file
   */
  public void checkInitializers(CContext context) throws PositionedError {
    self = createContext(context);

    if (statInit != null) {
      statInit.checkInitializer(self);
      staticInfo = self.getVariableInfo();
    }

    if (getCClass().getSuperClass() != null) {
      check(context,
	    !(getCClass().descendsFrom(getCClass().getSuperClass()) &&
	      getCClass().getSuperClass().descendsFrom(getCClass())),
	    KjcMessages.CLASS_CIRCULARITY,
	    ident);
    }

    // Check inners
    for (int i = inners.length - 1; i >= 0 ; i--) {
      inners[i].checkInitializers(self);
    }

    super.checkInitializers(context);
  }

  /**
   * checkTypeBody
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError	an error with reference to the source file
   */
  public void checkTypeBody(CContext context) throws PositionedError {
    try {
      CVariableInfo	instanceInfo = null;
      CVariableInfo[]	constructorsInfo;
      Vector		constructors = getDefaultConstructor() == null ? VectorCache.request(VectorCache.SMALL) : null;

      if (instanceInit != null) {
	instanceInit.checkInitializer(self);
	instanceInfo = self.getVariableInfo();
      }

      for (int i = fields.length - 1; i >= 0 ; i--) {
	((CSourceField)fields[i].getField()).setFullyDeclared(true);
      }

      for (int i = 0; i < inners.length; i++) {
	try {
	  inners[i].checkTypeBody(self);
	} catch (CBlockError e) {
	  context.reportTrouble(e);
	}
      }

      // First we compile constructors
      for (int i = methods.length - 1; i >= 0 ; i--) {
	try {
	  if (methods[i] instanceof JConstructorDeclaration) {
	    methods[i].checkStatement(self);
	    constructors.addElement(self.getVariableInfo());
	  }
	} catch (CBlockError e) {
	  context.reportTrouble(e);
	}
      }

      if (getDefaultConstructor() == null) {
	constructorsInfo = (CVariableInfo[])at.dms.util.Utils.vectorToArray(constructors, CVariableInfo.class);
	VectorCache.release(constructors);
      } else {
	getDefaultConstructor().checkStatement(self);
	constructorsInfo = new CVariableInfo[] {self.getVariableInfo()};
      }

      // Now we compile methods
      for (int i = methods.length - 1; i >= 0 ; i--) {
	try {
	  if (!(methods[i] instanceof JConstructorDeclaration)) {
	    methods[i].checkStatement(self);
	  }
	} catch (CBlockError e) {
	  context.reportTrouble(e);
	}
      }

      // Now we checks members
      for (int i = methods.length - 1; i >= 0 ; i--) {
	if (!((CSourceMethod)methods[i].getMethod()).isUsed() &&
	    !methods[i].getMethod().getIdent().equals(JAV_CONSTRUCTOR)) {
	  context.reportTrouble(new CWarning( methods[i].getTokenReference(), 
					     KjcMessages.UNUSED_PRIVATE_METHOD, 
					     methods[i].getMethod().getIdent()));
	}
      }
      for (int i = fields.length - 1; i >= 0 ; i--) {
	if (!((CSourceField)fields[i].getField()).isUsed()) {
	  context.reportTrouble(new CWarning(fields[i].getTokenReference(), 
					     KjcMessages.UNUSED_PRIVATE_FIELD, 
					     fields[i].getField().getIdent()));
	}
      }

      self.close(this, staticInfo, instanceInfo, constructorsInfo);

      super.checkTypeBody(context);
    } catch (UnpositionedError cue) {
      throw cue.addPosition(getTokenReference());
    }
    
    self = null;
  }

  /**
   * createContext
   * @param	parent		the parent context or null
   * @return	returns a CClassContext that represents this context
   * @exception	UnpositionedError	Error catched as soon as possible
   */
  public CClassContext createContext(CContext parent) {
    //return context.createContext(sourceClass);
    return new CClassContext(parent, sourceClass, this);
  }

  /**
   * check static initializers
   */
  private JConstructorDeclaration constructDefaultConstructor() {
    return new JConstructorDeclaration(getTokenReference(),
				       getCClass().isPublic() ? ACC_PUBLIC : 0,
				       ident,
				       JFormalParameter.EMPTY,
				       CClassType.EMPTY,
				       new JConstructorBlock(getTokenReference(), new JStatement[0]),
				       null,
				       null);
  }

  /**
   * Collects all initializers and builds a single method.
   * @param	isStatic	class or instance initializers ?
   */
  private JInitializerDeclaration constructInitializers(boolean isStatic) {
    Vector		elems = VectorCache.request(VectorCache.SMALL);
    boolean		needGen = false;

    for (int i = 0; i < body.length; i++) {
      if ((body[i] instanceof JClassBlock)
	  && (((JClassBlock)body[i]).isStaticInitializer() == isStatic)) {
	elems.addElement(body[i]);
	needGen = true;
      } else {
	if ((body[i] instanceof JFieldDeclaration)
	    && (((JFieldDeclaration)body[i]).getVariable().isStatic() == isStatic)) {
	  needGen |= ((JFieldDeclaration)body[i]).needInitialization();
	  elems.addElement(new JClassFieldDeclarator(getTokenReference(), (JFieldDeclaration)body[i]));
	}
      }
    }

    JStatement[]	stmts = (JStatement[])at.dms.util.Utils.vectorToArray(elems, JStatement.class);

    VectorCache.release(elems);
    if (stmts.length > 0) {
      return new JInitializerDeclaration(getTokenReference(),
					 new JClassBlock(getTokenReference(), false, stmts),
					 isStatic,
					 !needGen);
    } else {
      return null;
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);

    p.visitClassDeclaration(this,
			    getCClass().getModifiers(),
			    getCClass().getIdent(),
			    superName, //sourceClass.getSuperClass().getJavaName(),
			    interfaces,
			    body,
			    methods,
			    inners);
  }

  /**
   * Generate the code in pure java form
   * It is useful to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genInnerJavaCode(KjcPrettyPrinter p) {
    super.accept(p);

    p.visitInnerClassDeclaration(this, modifiers,
				 ident,
				 superName, //sourceClass.getSuperClass().getJavaName(),
				 interfaces,
				 inners,
				 body,
				 methods);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CVariableInfo		staticInfo;
  private	CClassContext		self;
  protected	String			superName;
}
