/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JInterfaceDeclaration.java,v 1.2 2002/03/08 21:54:31 mrobilla Exp $
 */

package at.dms.kjc;

import java.util.Vector;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.util.VectorCache;

/**
 * This class represents a java class in the syntax tree
 */
public class JInterfaceDeclaration extends JTypeDeclaration {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	context		the context in which this class is defined
   * @param	modifiers	the list of modifiers of this class
   * @param	shortName	the short name of this class
   * @param	interfaces	the name of this class's interfaces
   * @param	javadoc	is this interface deprecated
   */
  public JInterfaceDeclaration(TokenReference where,
			       int modifiers,
			       String shortName,
			       CClassType[] interfaces,
			       JFieldDeclaration[] fields,
			       JMethodDeclaration[] methods,
			       JTypeDeclaration[] inners,
			       JPhylum[] initializers,
			       JavadocComment javadoc,
			       JavaStyleComment[] comments) {
    super(where, modifiers | ACC_INTERFACE | ACC_ABSTRACT,
	  shortName, interfaces, fields, methods, inners, initializers, javadoc, comments);
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

    /**
     * Second pass (quick), check interface looks good
     * Exceptions are not allowed here, this pass is just a tuning
     * pass in order to create informations about exported elements
     * such as Classes, Interfaces, Methods, Constructors and Fields
     * @return true iff sub tree is correct enought to check code
     * @exception	PositionedError	an error with reference to the source file
     */
    public void checkInterface(CContext context) throws PositionedError
    {
	statInit = constructStaticInitializers();
	super.checkInterface(context, CStdType.Object);
    }

  /**
   * Check that initializers are correct
   * @exception	PositionedError	an error with reference to the source file
   */
  public void checkInitializers(CContext context) throws PositionedError {
    CInterfaceContext self = (CInterfaceContext)createContext(context);

    if (statInit != null) {
      statInit.checkInitializer(self);
    }

    // Check inners
    for (int i = inners.length - 1; i >= 0 ; i--) {
      inners[i].checkInitializers(self);
    }

    super.checkInitializers(context);
  }

  /**
   * checkTypeBody
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError	an error with reference to the source file
   */
  public void checkTypeBody(CContext context) throws PositionedError {
    CInterfaceContext self = (CInterfaceContext)createContext(context);

    for (int i = 0; i < inners.length; i++) {
      try {
	inners[i].checkTypeBody(self);
      } catch (CBlockError e) {
	context.reportTrouble(e);
      }
    }

    for (int i = methods.length - 1; i >= 0 ; i--) {
      try {
	methods[i].checkStatement(self);
      } catch (PositionedError ce) {
	context.reportTrouble(ce);
      }
    }
    
    super.checkTypeBody(context);
  }

  /**
   * createContext
   * @param	parent		the parent context or null
   * @return	returns a CClassContext that represents this context
   * @exception	UnpositionedError	Error catched as soon as possible
   */
  public CClassContext createContext(CContext parent) {
    return new CInterfaceContext(parent, sourceClass);
  }

  /**
   * check static initializers
   */
  private JInitializerDeclaration constructStaticInitializers() {
    // collect all static initializers and build a method
    Vector	elems = VectorCache.request(VectorCache.SMALL);
    boolean	needGen = false;
    for (int i = 0; i < body.length; i++) {
      if (body[i] instanceof JClassBlock) {
	elems.addElement(body[i]);
	needGen = true;
      } else {
	if (body[i] instanceof JFieldDeclaration) {
	  needGen |= ((JFieldDeclaration)body[i]).needInitialization();
	  elems.addElement(new JClassFieldDeclarator(getTokenReference(), (JFieldDeclaration)body[i]));
	}
      }
    }

    JStatement[]	stmts = (JStatement[])at.dms.util.Utils.vectorToArray(elems, JStatement.class);

    VectorCache.release(elems);
    if (stmts.length > 0) {
      return new JInitializerDeclaration(getTokenReference(),
					 new JClassBlock(getTokenReference(), false, stmts),
					 true,
					 !needGen);
    } else {
      return null;
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);
    p.visitInterfaceDeclaration(this,
				getCClass().getModifiers(),
				sourceClass.getIdent(),
				interfaces,
				body,
				methods);
  }
}
