/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLabeledStatement.java,v 1.2 2002/03/08 21:54:31 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.6: Labeled Statement
 *
 * Statements may have label prefixes.
 */
public class JLabeledStatement extends JStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	label		the label of the enclosing labeled statement
   * @param	body		the contained statements
   * @param	comments	comments in the source text
   */
  public JLabeledStatement(TokenReference where, String label, JStatement stmt, JavaStyleComment[] comments) {
    super(where, comments);

    this.label = label;
    this.stmt = stmt;
    this.isBreakTarget = false;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Returns the label of this statement.
   */
  /*package*/ String getLabel() {
    return label;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    check(context,
	  context.getLabeledStatement(label) == null,
	  KjcMessages.LABEL_ALREADY_EXISTS, label);

    CLabeledContext	self = new CLabeledContext(context, this);
    CBodyContext	body = new CBodyContext(self);

    stmt.checkBody(body);

    body.close(getTokenReference());
    self.close(getTokenReference());

    if (isBreakTarget) {
      context.setReachable(true);
    } else {
      context.reportTrouble(new CWarning(getTokenReference(), KjcMessages.UNUSED_LABEL, label));
    }
  }

  // ----------------------------------------------------------------------
  // BREAK/CONTINUE HANDLING
  // ----------------------------------------------------------------------

  /**
   * Returns true iff this statement is a possible target for a break statement.
   */
  public boolean acceptsBreak() {
    return true;
  }

  /**
   * Returns true iff this statement is a possible target for a continue statement.
   */
  public boolean acceptsContinue() {
    return stmt.acceptsContinue();
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    super.accept(p);
    p.visitLabeledStatement(this, label, stmt);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    stmt.genCode(code);
    code.plantLabel(endLabel);

    endLabel = null;
  }

  /**
   * Prevent statement that there is at least one break
   */
  public void addBreak(CBodyContext context) {
    isBreakTarget = true;
    if (stmt.acceptsBreak()) {
      stmt.addBreak(context);
    }
  }

  /**
   * Return the end of this block (for break statement)
   */
  public CodeLabel getBreakLabel() {
    if (stmt.acceptsBreak()) {
      return stmt.getBreakLabel();
    } else {
      return endLabel;
    }
  }

  /**
   * Return the beginning of this block (for continue statement)
   */
  public CodeLabel getContinueLabel() {
    return stmt.getContinueLabel();
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private String		label;
  private JStatement		stmt;
  private CodeLabel		endLabel = new CodeLabel();
  private boolean		isBreakTarget;
  private boolean		isContinueTarget;
}
