/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLocalVariableExpression.java,v 1.2 2002/03/08 21:54:32 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Root class for all expressions
 */
public class JLocalVariableExpression extends JExpression implements CInitializable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where the line of this node in the source code
   */
  public JLocalVariableExpression(TokenReference where, JLocalVariable variable) {
    super(where);
    this.variable = variable;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Returns true if this field accept assignment
   */
  public boolean isLValue(CExpressionContext context) {
    return !variable.isFinal() || !isInitialized(context);
  }

  /**
   * Returns true if this field is already initialized
   */
  public boolean isInitialized(CExpressionContext context) {
    return CVariableInfo.isInitialized(context.getBodyContext().getVariableInfo(variable.getIndex()));
  }

  /**
   * Returns true if this field is already initialized
   */
  public void initialize(CBodyContext context) {
    context.setVariableInfo(variable.getIndex(), CVariableInfo.initialize());
  }

  /**
   * Returns the position of this variable in the sets of local vars
   */
  public int getPosition() {
    return variable.getPosition();
  }

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return variable.getType();
  }

  public String getIdent() {
    return variable.getIdent();
  }  

  /**
   * Returns true iff this expression is constant.
   */
  public boolean isConstant() {
    return variable.isConstant();
  }

  /**
   * Returns the literal value of this field
   */
  public JLiteral getLiteral() {
    return (JLiteral)variable.getValue();
  }  

  public JLocalVariable getVariable() {
    return variable;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    if (!context.isLeftSide() || !context.discardValue()) {
      variable.setUsed();
    } 
    if (context.isLeftSide()) {
      variable.setAssigned(getTokenReference(), context.getBodyContext());
    }

    check(context, CVariableInfo.isInitialized(context.getBodyContext().getVariableInfo(variable.getIndex())) ||
	  (context.isLeftSide() && context.discardValue()),
	  KjcMessages.UNINITIALIZED_LOCAL_VARIABLE, variable.getIdent());

    if (variable.isConstant() && !context.isLeftSide()) {
      return variable.getValue();
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitLocalVariableExpression(this, variable.getIdent());
  }

  public boolean equals(Object o) {
    return (o instanceof JLocalVariableExpression) &&
      variable.equals(((JLocalVariableExpression)o).variable);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (!code.discardValue()) {
      variable.genLoad(code);
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genStartStoreCode(CodeSequence code) {
    // nothing to do here
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genEndStoreCode(CodeSequence code, boolean discardValue) {
    if (!discardValue) {
      int	opcode;

      if (getType().getSize() == 2) {
	opcode = opc_dup2;
      } else {
	opcode = opc_dup;
      }
      code.plantInstruction(new NoArgInstruction(opcode));
    }
    variable.genStore(code);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  JLocalVariable	variable;
}
