/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMinusExpression.java,v 1.2 2002/03/08 21:54:33 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements '-' specific operations
 * Plus operand may be String, numbers
 */
public class JMinusExpression extends JBinaryArithmeticExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JMinusExpression(TokenReference where,
			  JExpression left,
			  JExpression right) {
    super(where, left, right);
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param	context		the actual context of analyse
   * @return	a pure java expression including promote node
   * @exception	UnpositionedError Error catched as soon as possible
   */
  public JExpression checkBinaryBody(CExpressionContext context) throws PositionedError {
    try {
      type = computeType(left.getType(), right.getType());
    } catch (UnpositionedError e) {
      throw e.addPosition(getTokenReference());
    }
    
    if (left.isConstant() && right.isConstant()) {
      return constantFolding();
    } else {
      return this;
    }
  }

  /**
   * compute the type of this expression according to operands
   * @param	leftType		the type of left operand
   * @param	rightType		the type of right operand
   * @return	the type computed for this binary operation
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public static CType computeType(CType	leftType, CType rightType) throws UnpositionedError {
    if (leftType.isNumeric() && rightType.isNumeric()) {
      return CNumericType.binaryPromote(leftType, rightType);
    }
    throw new UnpositionedError(KjcMessages.MINUS_BADTYPE, leftType, rightType);
  }

  /**
   * Computes the result of the operation at compile-time (JLS 15.27).
   * @param	left		the left value
   * @param	right		the right value
   * @return	a literal resulting of an operation over two literals
   */
  public JExpression constantFolding() {
    if (left.getType() == CStdType.Long) {
      long		result;
      
      result = compute(left.getNumberLiteral().numberValue().longValue(), right.getNumberLiteral().numberValue().longValue());
      return new JOrdinalLiteral(getTokenReference(), new Long(result), CStdType.Long);
    } else if (left.getType() == CStdType.Float) {
      float		result;
      
      result = compute(left.getNumberLiteral().numberValue().floatValue(), right.getNumberLiteral().numberValue().floatValue());
      return new JRealLiteral(getTokenReference(), new Float(result), CStdType.Float);
    } else if (left.getType() == CStdType.Double) {
      double		result;
      
      result = compute(left.getNumberLiteral().numberValue().doubleValue(), right.getNumberLiteral().numberValue().doubleValue());
      return new JRealLiteral(getTokenReference(), new Double(result), CStdType.Double);
    } else if (left.getType().implicitlyCastTo(CStdType.Integer)) {
      int		result;

      result = compute(left.getNumberLiteral().numberValue().intValue(), right.getNumberLiteral().numberValue().intValue());
      return new JOrdinalLiteral(getTokenReference(), new Integer(result), CStdType.Integer);
    } else {
      fail();
      return null;
    }
  }

  /**
   * Computes the result of the operation at compile-time (JLS 15.27).
   * @param	left		the first operand
   * @param	right		the seconds operand
   * @return	the result of the operation
   */
  public int compute(int left, int right) {
    return left - right;
  }

  /**
   * Computes the result of the operation at compile-time (JLS 15.27).
   * @param	left		the first operand
   * @param	right		the seconds operand
   * @return	the result of the operation
   */
  public long compute(long left, long right) {
    return left - right;
  }

  /**
   * Computes the result of the operation at compile-time (JLS 15.27).
   * @param	left		the first operand
   * @param	right		the seconds operand
   * @return	the result of the operation
   */
  public float compute(float left, float right) {
    return left - right;
  }

  /**
   * Computes the result of the operation at compile-time (JLS 15.27).
   * @param	left		the first operand
   * @param	right		the seconds operand
   * @return	the result of the operation
   */
  public double compute(double left, double right) {
    return left - right;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitBinaryExpression(this, "-", left, right);
  }

  /**
   * @param	type		the type of result
   * @return	the type of opcode for this operation
   */
  public static int getOpcode(CType type) {
    switch (type.getTypeID()) {
    case TID_FLOAT:
      return opc_fsub;
    case TID_LONG:
      return opc_lsub;
    case TID_DOUBLE:
      return opc_dsub;
    default:
      return opc_isub;
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    left.genCode(code);
    right.genCode(code);

    code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
  }
}
