/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JThisExpression.java,v 1.2 2002/03/08 21:54:38 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * A 'this' expression
 */
public class JThisExpression extends JExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where the line of this node in the source code
   */
  public JThisExpression(TokenReference where) {
    super(where);
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	self		the class onto this suffix is applied
   */
  public JThisExpression(TokenReference where, CClass self) {
    this(where);
    this.self = self;
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	prefix		the left expression like t.this
   */
  public JThisExpression(TokenReference where, JExpression prefix) {
    this(where);
    this.prefix = prefix;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return self.getType();
  }

  /**
   * @return is this expression a lvalue ?
   */
  public boolean isLValue(CExpressionContext context) {
    return false;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(final CExpressionContext context) throws PositionedError {
    if (prefix != null) {
      prefix = prefix.checkExpression(context);
      check(context, prefix.getType().isClassType(), KjcMessages.THIS_BADACCESS);
      self = prefix.getType().getCClass();
    } else if (self == null) {
      self = context.getClassContext().getCClass();
    }

    if (!context.getClassContext().getCClass().descendsFrom(self)) {
      // access to outer class
      CClass		clazz = context.getClassContext().getCClass();
      JExpression	expr = null;
      CClassContext	classContext = context.getClassContext();

      while (!clazz.descendsFrom(self)) {
	check(context, 
	      !classContext.getTypeDeclaration().getCClass().isStatic(),
	      KjcMessages.THIS_INVALID_OUTER);
	classContext.getTypeDeclaration().addOuterThis();
	classContext = classContext.getParentContext().getClassContext();
	
	if (expr == null) {
	  if (context.getMethodContext() instanceof CConstructorContext) {
	    JGeneratedLocalVariable local = new JGeneratedLocalVariable(null, 0, getType(), "toto", null) {
		/**
		 * @return the local index in context variable table
		 */
		public int getPosition() {
		  return context.getMethodContext().getCMethod().getParameters().length + 1 /*this*/;
		}
	      };
	    
	    return new JLocalVariableExpression(getTokenReference(), local) {
		public JExpression checkExpression(CExpressionContext ctxt) {
		  // already checked
		  return this;
		}
	      };
	  } else {
	    expr = new JClassFieldExpression(getTokenReference(), new JThisExpression(getTokenReference()), JAV_OUTER_THIS);
	  }
	} else {
	  expr = new JClassFieldExpression(getTokenReference(), expr, JAV_OUTER_THIS);
	}

	expr = expr.checkExpression(context);
	clazz = ((CClassType)expr.getType()).getCClass();
      }

      if (prefix != null) {
	check(context, expr.getType().equals(prefix.getType()) ||
	      /*May be it is an innerclass with the same name, therefore the prefix name has been
	       wrongly assigned to the outer. So we compare the names instead:*/
	      ((CClassType)expr.getType()).getCClass().getIdent().equals(((CClassType)prefix.getType()).getCClass().getIdent()),
	      KjcMessages.THIS_INVALID_OUTER, prefix.getType());
      }
      return expr;
    }

    check(context, !context.getMethodContext().getCMethod().isStatic(), KjcMessages.BAD_THIS_STATIC);

    return this;
  }

  public boolean equals(Object o) {
    return o instanceof JThisExpression &&
      self.equals(((JThisExpression)o).self);
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitThisExpression(this, prefix);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    if (!code.discardValue()) {
      code.setLineNumber(getTokenReference().getLine());

      code.plantLoadThis();
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CClass		self;
  private	JExpression	prefix;
}
