/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTryCatchStatement.java,v 1.2 2002/03/08 21:54:38 mrobilla Exp $
 */

package at.dms.kjc;

import java.util.Enumeration;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.18: Try Statement
 *
 * A try statement executes a block.
 * If a value is thrown and the try statement has one or more catch
 * clauses that can catch it, then control will be transferred to the
 * first such catch clause.
 * If the try statement has a finally clause, then another block of code
 * is executed, no matter whether the try block completes normally or abruptly,
 * and no matter whether a catch clause is first given control.
 *
 * In this implementation, the Try Statement is split into a Try-Catch Statement
 * and a Try-Finally Statement. A Try Statement where both catch and finally
 * clauses are present is rewritten as a Try-Catch Statement enclosed in a
 * Try-Finally Statement.
 */
public class JTryCatchStatement extends JStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where			the line of this node in the source code
   * @param	tryClause		the body
   * @param	catchClauses		a vector of catch clause
   */
  public JTryCatchStatement(TokenReference where,
			    JBlock tryClause,
			    JCatchClause[] catchClauses,
			    JavaStyleComment[] comments)
  {
    super(where, comments);

    this.tryClause = tryClause;
    this.catchClauses = catchClauses;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    CBodyContext	unchecked = context.cloneContext();
    CTryContext		tryContext = new CTryContext(context);
    CBodyContext	inside = new CBodyContext(tryContext);
    int			count = 0;
    boolean		runtimeCatched = false;
    boolean		errorCatched = false;
    boolean		isReachable = false;

    tryClause.checkStatement(inside);
    inside.close(getTokenReference());
    if (inside.isReachable()) {
      count ++;
      isReachable = true;
    }

    for (int i = 0; i < catchClauses.length; i++) {
      catchClauses[i].checkInterface(context);
    }
    
    Enumeration		enum = tryContext.getThrowables().elements();
  loop:
    while (enum.hasMoreElements()) {
      CThrowableInfo	info = (CThrowableInfo)enum.nextElement();
      CClassType	type = info.getThrowable();

      for (int i = 0; i < catchClauses.length; i++) {
	if (type.isAssignableTo(catchClauses[i].getType()) || 
	    catchClauses[i].getType().isAssignableTo(type)) {
	  if (!catchClauses[i].isChecked()) {
	    isReachable |= checkClause(context, info.getContext(), catchClauses[i], count++ == 0);
	  }
	  if (!catchClauses[i].getType().equals(type) &&
	      catchClauses[i].getType().isAssignableTo(type)) {
	    continue;
	  } else {
	    continue loop;
	  }
	}
      }
      // not catched
      context.addThrowable(info);
    }
    
    // CHEKC ORDER $$$ WRONG WAY
    for (int i = 0; i < catchClauses.length; i++) {
      for (int j = i + 1; j < catchClauses.length; j++) {
	if (catchClauses[j].getType().isAssignableTo(catchClauses[i].getType())) {
	  throw new CBlockError(catchClauses[j].getTokenReference(), KjcMessages.STATEMENT_UNREACHABLE);
	}
      }
    }

    for (int i = 0; i < catchClauses.length; i++) {
      if (!catchClauses[i].isChecked()) {
	CClassType	type = catchClauses[i].getType();
	if (type.isCheckedException() && !type.equals(CStdType.Throwable) && !type.equals(CStdType.Exception)) {
	  context.reportTrouble(new CBlockError(catchClauses[i].getTokenReference(), KjcMessages.STATEMENT_UNREACHABLE));
	} else if (type.equals(CStdType.RuntimeException) && !runtimeCatched) {
	  isReachable |= checkClause(context, unchecked, catchClauses[i], count++ == 0);
	  runtimeCatched = true;
	} else if (type.equals(CStdType.Error) && !errorCatched) {
	  isReachable |= checkClause(context, unchecked, catchClauses[i], count++ == 0);
	  errorCatched = true;
	} else if (!type.equals(CStdType.Error) && !type.equals(CStdType.RuntimeException)) {
	  isReachable |= checkClause(context, unchecked, catchClauses[i], count++ == 0);
	} else {
	  context.reportTrouble(new CBlockError(catchClauses[i].getTokenReference(), KjcMessages.STATEMENT_UNREACHABLE)); 
	  // runtime exception more than one time	  
	}
      }
    }

    tryContext.setThrowables(null);
    tryContext.close(getTokenReference());
    context.setReachable(isReachable);
    check(context, catchClauses.length > 0, KjcMessages.TRY_NOCATCH);
  }

  private boolean checkClause(CBodyContext context, CBodyContext body, JCatchClause clause, boolean first) throws PositionedError {
    clause.checkBody(body);
	  
    // Reachable => add infos
    if (body.isReachable()) {
      if (first) {
	context.adopt(body);
      } else {
	context.merge(body);
      }
    }
    
    // Add throwables
    Enumeration		enum2 = body.getThrowables().elements();
    while (enum2.hasMoreElements()) {
      context.addThrowable((CThrowableInfo)enum2.nextElement());
    }

    return body.isReachable();
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitTryCatchStatement(this, tryClause, catchClauses);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    CodeLabel		nextLabel = new CodeLabel();

    int		startPC = code.getPC();
    tryClause.genCode(code);			//		TRY CODE
    int		endPC = code.getPC();
    code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
    for (int i = 0; i < catchClauses.length; i++) {
      catchClauses[i].genCode(code);
      catchClauses[i].genExceptionHandler(code, startPC, endPC);
      code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
    }
    code.plantLabel(nextLabel);			//	next:	...
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private JBlock		tryClause;
  private JCatchClause[]	catchClauses;
}
