/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JUnaryExpression.java,v 1.2 2002/03/08 21:54:39 mrobilla Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents unary expressions.
 */
public class JUnaryExpression extends JExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	oper		the operator
   * @param	expr		the operand
   */
  public JUnaryExpression(TokenReference where, int oper, JExpression expr) {
    super(where);
    this.oper = oper;
    this.expr = expr;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    if ((expr instanceof JOrdinalLiteral) && oper == OPE_MINUS) {
      ((JOrdinalLiteral)expr).setSignToMinus();
      oper = OPE_PLUS;
    }
    expr = expr.checkExpression(context);

    switch (oper) {
    case OPE_PLUS:
    case OPE_MINUS:
      check(context, expr.getType().isNumeric(), KjcMessages.UNARY_BADTYPE_PM, expr.getType());
      type = CNumericType.unaryPromote(expr.getType());

      if (expr.isConstant()) {
	if (oper == OPE_PLUS) {
	  return expr;
	}

	JNumberLiteral	lit = expr.getNumberLiteral();
	if (type == CStdType.Double) {
	  lit = new JRealLiteral(getTokenReference(),
				 new Double(-lit.numberValue().doubleValue()),
				 CStdType.Double);
	} else if (type == CStdType.Float) {
	  lit = new JRealLiteral(getTokenReference(),
				 new Float(-lit.numberValue().floatValue()),
				 CStdType.Float);
	} else if (type == CStdType.Long) {
	  lit = new JOrdinalLiteral(getTokenReference(),
				    new Long(-lit.numberValue().longValue()),
				    CStdType.Long);
	} else {
	  lit = new JOrdinalLiteral(getTokenReference(),
				    new Integer(-lit.numberValue().intValue()),
				    CStdType.Integer);
	}

	return lit;
      }

      break;

    case OPE_BNOT:
      check(context, expr.getType().isOrdinal(), KjcMessages.UNARY_BADTYPE_BNOT, expr.getType());
      type = CNumericType.unaryPromote(expr.getType());

      if (expr.isConstant()) {
	JOrdinalLiteral	lit = expr.getOrdinalLiteral();
	if (type == CStdType.Long) {
	  lit = new JOrdinalLiteral(getTokenReference(),
				    new Long(~lit.numberValue().longValue()),
				    CStdType.Long);
	} else {
	  lit = new JOrdinalLiteral(getTokenReference(),
				    new Integer(~lit.numberValue().intValue()),
				    CStdType.Integer);
	}
	return lit;
      }
      break;

    case OPE_LNOT:
      check(context, expr.getType() == CStdType.Boolean, KjcMessages.UNARY_BADTYPE_LNOT, expr.getType());
      type = CStdType.Boolean;
      if (expr.isConstant()) {
	return new JBooleanLiteral(getTokenReference(), !expr.getBooleanLiteral().booleanValue());
      }
      break;

    default:
      fail();
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Accepts the specified visitor
   * @param	p		the visitor
   */
  public void accept(KjcVisitor p) {
    p.visitUnaryExpression(this, oper, expr);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    switch (oper) {
    case OPE_PLUS:
      expr.genCode(code);
      break;

    case OPE_MINUS:
      expr.genCode(code);
      switch (type.getTypeID()) {
      case TID_FLOAT:
	code.plantInstruction(new NoArgInstruction(opc_fneg));
	break;
      case TID_LONG:
	code.plantInstruction(new NoArgInstruction(opc_lneg));
	break;
      case TID_DOUBLE:
	code.plantInstruction(new NoArgInstruction(opc_dneg));
	break;
      default:
	code.plantInstruction(new NoArgInstruction(opc_ineg));
      }
      break;

    case OPE_LNOT:
      expr.genCode(code);
      code.plantInstruction(new PushLiteralInstruction(1));
      code.plantInstruction(new NoArgInstruction(opc_ixor));
      break;

    case OPE_BNOT:
      expr.genCode(code);
      if (type == CStdType.Long) {
	code.plantInstruction(new PushLiteralInstruction((long)-1));
	code.plantInstruction(new NoArgInstruction(opc_lxor));
      } else {
	code.plantInstruction(new PushLiteralInstruction((int)-1));
	code.plantInstruction(new NoArgInstruction(opc_ixor));
      }
      break;

    default:
      fail();
    }
  }

  /**
   * Generates a sequence of bytescodes to branch on a label
   * This method helps to handle heavy optimizables conditions
   * @param	code		the code list
   */
  public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
    assert(oper == OPE_LNOT);

    expr.genBranch(!cond, code, label);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected int			oper;
  protected JExpression		expr;
  protected CType		type;
}
