/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package at.dms.kjc;

import at.dms.compiler.JavaStyleComment;
import at.dms.compiler.JavadocComment;
import java.util.Vector;
import java.util.Stack;
import ca.ubc.cs.jex.JexFileCollection;
import ca.ubc.cs.jex.JexFile;
import ca.ubc.cs.jex.JexMethod;
import ca.ubc.cs.jex.AnonymousJexFile;

/**
  * Collects all the class declarations in the file and creates
  * JexFiles for them initialized with all methods.
  */
public class TypeDeclarationCollectorVisitor implements KjcVisitor
{
    private Vector aDecls = new Vector();
    private JexFileCollection aJexFiles = new JexFileCollection();
    private boolean aNextAnonymous = false;
    private Stack   aCurrentJexFile = new Stack();

    public JexFileCollection getJexFileCollection()
    { return aJexFiles; }

    // ----------------------------------------------------------------------
    // COMPILATION UNIT
    // ----------------------------------------------------------------------
    
    /**
     * visits a compilation unit
     */
    public void visitCompilationUnit(JCompilationUnit self,
			      JPackageName packageName,
			      JPackageImport[] importedPackages,
			      JClassImport[] importedClasses,
			      JTypeDeclaration[] typeDeclarations)
    {
	if( typeDeclarations != null )
	    for( int i = 0; i < typeDeclarations.length; i++ )
		typeDeclarations[i].accept(this);
    }
    
    // ----------------------------------------------------------------------
    // TYPE DECLARATION
    // ----------------------------------------------------------------------
    
    /**
     * visits a class declaration
     */
   public  void visitClassDeclaration(JClassDeclaration self,
			       int modifiers,
			       String ident,
			       String superName,
			       CClassType[] interfaces,
			       JPhylum[] body,
			       JMethodDeclaration[] methods,
			       JTypeDeclaration[] decls)
    {
	// Create a new Jex file for this class
	JexFile lJexFile = null;
	if( aNextAnonymous )
	{
	    aNextAnonymous = false;
	    lJexFile = new AnonymousJexFile( self.getCClass() );
	}
	else
	    lJexFile = new JexFile( self.getCClass() );
	
	// Add the Jex file to the collection
	aJexFiles.addJexFile( self.getCClass().getQualifiedName(), lJexFile );
	aJexFiles.addClassDecl( self );
	
	// Stick the Jex file on the stack for processing
	aCurrentJexFile.push( lJexFile );

	JFieldDeclaration[] lFields = self.getFields();
	if( lFields != null )
	    for( int i = 0; i < lFields.length; i++ )
		lFields[i].accept( this );

        JConstructorDeclaration lDefault = self.getDefaultConstructor();
	if( lDefault != null )
	    lDefault.accept( this );

	JInitializerDeclaration lInstanceInit = self.instanceInit;
	if( lInstanceInit != null )
	    lInstanceInit.accept( this );

	if( methods != null )
	    for( int i = 0; i < methods.length; i++ )
		methods[i].accept( this );
	
	if( decls != null )
	    for( int i = 0; i < decls.length; i++ )
		decls[i].accept( this );

	aCurrentJexFile.pop();
    }

    /**
     * visits a class body
     */
    public void visitClassBody(JTypeDeclaration[] decls,
			JMethodDeclaration[] methods,
			JPhylum[] body){}
    
    /**
     * visits a class declaration
     */
    public void visitInnerClassDeclaration(JClassDeclaration self,
				    int modifiers,
				    String ident,
				    String superName,
				    CClassType[] interfaces,
				    JTypeDeclaration[] decls,
				    JPhylum[] body,
				    JMethodDeclaration[] methods){}

  /**
   * visits an interface declaration
   */
  public void visitInterfaceDeclaration(JInterfaceDeclaration self,
				 int modifiers,
				 String ident,
				 CClassType[] interfaces,
				 JPhylum[] body,
				 JMethodDeclaration[] methods)
    {
	// We don't add the interface declaration to the Jex File
	// collection because it contains no method body and
	// no Jex file should be generated for it.  Nevertheless,
	// we process the members for inner class declarations.
	// To make sure the methods don't get added to the JexFile
	// on the stack, we put a marker.

	aCurrentJexFile.push( new Integer(1) );

	JFieldDeclaration[] lFields = self.getFields();
	if( lFields != null )
	    for( int i = 0; i < lFields.length; i++ )
		lFields[i].accept( this );
        
	JConstructorDeclaration lDefault = self.getDefaultConstructor();
	if( lDefault != null )
	    lDefault.accept( this );
	
	JInitializerDeclaration lInstanceInit = self.instanceInit;
	if( lInstanceInit != null )
	    lInstanceInit.accept( this );
	
	if( methods != null )
	    for( int i = 0; i < methods.length; i++ )
		methods[i].accept( this );

	JTypeDeclaration[] lInners = self.inners;
	if( lInners != null )
	    for( int i = 0; i < lInners.length; i++ )
		lInners[i].accept( this );

	aCurrentJexFile.pop();
    }
    
    // ----------------------------------------------------------------------
    // METHODS AND FIELDS
    // ----------------------------------------------------------------------
    
    /**
     * visits a field declaration
     */
    public void visitFieldDeclaration(JFieldDeclaration self,
			       int modifiers,
			       CType type,
			       String ident,
			       JExpression expr)
    {
	if( expr != null )
	    expr.accept( this );
    }

    /**
     * visits a method declaration
     */
    public void visitMethodDeclaration(JMethodDeclaration self,
				int modifiers,
				CType returnType,
				String ident,
				JFormalParameter[] parameters,
				CClassType[] exceptions,
				JBlock body)
    {
	// If we're not in an interface, add the JexMethod
	if( (aCurrentJexFile.peek() instanceof Integer) == false )
	{
	    JexMethod lMethod = new JexMethod( JexVisitor.makeMethodSignature( ident, parameters), exceptions );
	    ((JexFile)aCurrentJexFile.peek()).addJexMethod( lMethod );
	}

	if( body != null )
	    body.accept( this );
    }
    
    /**
     * visits a method declaration
     */
    public void visitConstructorDeclaration(JConstructorDeclaration self,
				     int modifiers,
				     String ident,
				     JFormalParameter[] parameters,
				     CClassType[] exceptions,
				     JBlock body)
    {
	// If we're not in an interface (and we shouldn't!), add the JexMethod
	if( (aCurrentJexFile.peek() instanceof Integer) == false )
	{
	    JexMethod lMethod = new JexMethod( JexVisitor.makeConstructorSignature( parameters ), exceptions );
	    ((JexFile)aCurrentJexFile.peek()).addJexMethod( lMethod );
	}
	if( body != null )
	    body.accept( this );
    }
    
    // ----------------------------------------------------------------------
    // STATEMENTS
    // ----------------------------------------------------------------------
    
    /**
     * visits a while statement
     */
    public void visitWhileStatement(JWhileStatement self,
			     JExpression cond,
			     JStatement body)
    {
	if( cond != null ) cond.accept( this );
	if( body != null ) body.accept( this );
    }
    
    /**
   * visits a variable declaration statement
   */
    public void visitVariableDeclarationStatement(JVariableDeclarationStatement self,
					 JVariableDefinition[] vars)
    {
	if( vars != null )
	    for( int i = 0; i < vars.length; i++ )
		vars[i].accept(this);
    }

  /**
   * visits a variable declaration statement
   */
  public void visitVariableDefinition(JVariableDefinition self,
			       int modifiers,
			       CType type,
			       String ident,
			       JExpression expr)
    {
	if( expr != null )
	    expr.accept( this );
    }
    
  /**
   * visits a try-catch statement
   */
    public void visitTryCatchStatement(JTryCatchStatement self,
				JBlock tryClause,
				JCatchClause[] catchClauses)
    {
	if( tryClause != null ) tryClause.accept(this);
	if( catchClauses != null ) 
	    for( int i = 0; i < catchClauses.length; i++ )
		catchClauses[i].accept(this);
    }

  /**
   * visits a try-finally statement
   */
  public void visitTryFinallyStatement(JTryFinallyStatement self,
				JBlock tryClause,
				JBlock finallyClause)
    {
	if( tryClause != null ) tryClause.accept(this);
	if( finallyClause != null ) finallyClause.accept(this);
    }

    /**
     * visits a throw statement
     */
    public void visitThrowStatement(JThrowStatement self,
			     JExpression expr)
    {
	if( expr != null )
	    expr.accept(this);
    }

  /**
   * visits a synchronized statement
   */
  public void visitSynchronizedStatement(JSynchronizedStatement self,
				  JExpression cond,
				  JStatement body)
    {
	if( cond != null ) cond.accept(this);
	if( body != null ) body.accept(this);
    }

  /**
   * visits a switch statement
   */
  public void visitSwitchStatement(JSwitchStatement self,
			    JExpression expr, 
			    JSwitchGroup[] body)
    {
	if( expr != null ) expr.accept(this);
	if( body != null )
	    for( int i = 0; i < body.length; i++ )
		body[i].accept(this);
    }

  /**
   * visits a return statement
   */
  public void visitReturnStatement(JReturnStatement self,
			    JExpression expr)
    {
	if( expr != null ) expr.accept(this);
    }

  /**
   * visits a labeled statement
   */
  public void visitLabeledStatement(JLabeledStatement self,
			     String label,
			     JStatement stmt)
    {
	if( stmt != null ) stmt.accept(this);
    }

  /**
   * visits a if statement
   */
  public void visitIfStatement(JIfStatement self,
			JExpression cond,
			JStatement thenClause,
			JStatement elseClause)
    {
	if( cond != null ) cond.accept( this );
	if( thenClause != null ) thenClause.accept(this);
	if( elseClause != null ) elseClause.accept(this);
    }

  /**
   * visits a for statement
   */
  public void visitForStatement(JForStatement self,
			 JStatement init,
			 JExpression cond,
			 JStatement incr,
			 JStatement body)
    {
	if( init != null ) init.accept(this);
	if( cond != null ) cond.accept(this);
	if( incr != null ) incr.accept(this);
	if( body != null ) body.accept(this);
    }

  /**
   * visits a compound statement
   */
  public void visitCompoundStatement(JCompoundStatement self,
			      JStatement[] body)
    {
	if( body != null )
	    for( int i = 0; i < body.length; i++ )
		body[i].accept(this);
    }

  /**
   * visits an expression statement
   */
  public void visitExpressionStatement(JExpressionStatement self,
				JExpression expr)
    {
	if( expr != null )
	    expr.accept( this );
    }

  /**
   * visits an expression list statement
   */
  public void visitExpressionListStatement(JExpressionListStatement self,
				    JExpression[] expr)
    {
	if( expr != null )
	    for( int i = 0; i < expr.length; i++ )
		expr[i].accept(this);
    }

  /**
   * visits a empty statement
   */
    public void visitEmptyStatement(JEmptyStatement self){}

  /**
   * visits a do statement
   */
  public void visitDoStatement(JDoStatement self,
			JExpression cond,
			JStatement body)
    {
	if( cond != null ) cond.accept(this);
	if( body != null ) body.accept(this);
    }

  /**
   * visits a continue statement
   */
  public void visitContinueStatement(JContinueStatement self,
			      String label){}

  /**
   * visits a break statement
   */
  public void visitBreakStatement(JBreakStatement self,
      String label){}

  /**
   * visits an expression statement
   */
  public void visitBlockStatement(JBlock self,
			   JStatement[] body,
			   JavaStyleComment[] comments)
    {
	if( body != null )
	    for( int i = 0; i < body.length; i++ )
		body[i].accept(this);
    }

  /**
   * visits a type declaration statement
   */
  public void visitTypeDeclarationStatement(JTypeDeclarationStatement self,
				     JTypeDeclaration decl)
    {
	aNextAnonymous = true;
	if( decl != null ) decl.accept(this);
    }

  // ----------------------------------------------------------------------
  // EXPRESSION
  // ----------------------------------------------------------------------

  /**
   * visits an unary expression
   */
  public void visitUnaryExpression(JUnaryExpression self,
			    int oper,
			    JExpression expr)
    {
	if( expr != null )
	    expr.accept(this);
    }

  /**
   * visits a type name expression
   */
  public void visitTypeNameExpression(JTypeNameExpression self,
      CType type){}

  /**
   * visits a this expression
   */
  public void visitThisExpression(JThisExpression self,
			   JExpression prefix)
    {
	if( prefix != null )
	    prefix.accept(this);
    }

  /**
   * visits a super expression
   */
    public void visitSuperExpression(JSuperExpression self){}

  /**
   * visits a shift expression
   */
  public void visitShiftExpression(JShiftExpression self,
			    int oper, 
			    JExpression left, 
			    JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits a shift expressiona
   */
  public void visitRelationalExpression(JRelationalExpression self,
				 int oper, 
				 JExpression left, 
				 JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits a prefix expression
   */
  public void visitPrefixExpression(JPrefixExpression self,
			     int oper, 
			     JExpression expr)
    {
	if( expr != null )
	    expr.accept(this);
    }

  /**
   * visits a postfix expression
   */
  public void visitPostfixExpression(JPostfixExpression self,
			      int oper, 
			      JExpression expr)
    {
	if( expr!= null )
	    expr.accept(this);
    }

  /**
   * visits a parenthesed expression
   */
  public void visitParenthesedExpression(JParenthesedExpression self,
				  JExpression expr)
    {
	if( expr != null ) expr.accept(this);
    }

  /**
   * visits an object allocator expression
   */
  public void visitNewObjectExpression(JNewObjectExpression self,
				CClassType type, 
				JExpression[] params, 
				JClassDeclaration innerBody)
    {
	if( params!=null )
	    for( int i = 0; i < params.length; i++ )
		params[i].accept(this);
	
	if( innerBody != null )
        {
	    aNextAnonymous = true;
	    innerBody.accept(this);
	}
    }

  /**
   * visits an array allocator expression
   */
  public void visitNewArrayExpression(JNewArrayExpression self,
			       CType type, 
			       JArrayDimsAndInits dims)
    {
	if( dims !=null ) dims.accept(this);
    }

  /**
   * visits a name expression
   */
  public void visitNameExpression(JNameExpression self,
			   JExpression prefix, 
			   String ident)
    {
	if( prefix != null )
	    prefix.accept(this);
    }

  /**
   * visits an array allocator expression
   */
  public void visitBinaryExpression(JBinaryExpression self,
			     String oper,
			     JExpression left,
			     JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits a method call expression
   */
  public void visitMethodCallExpression(JMethodCallExpression self,
				 JExpression prefix,
				 String ident,
				 JExpression[] args)
    {
	if( prefix != null ) prefix.accept(this);
	if( args != null )
	    for( int i = 0; i < args.length; i++ )
		args[i].accept(this);
    }

  /**
   * visits a local variable expression
   */
  public void visitLocalVariableExpression(JLocalVariableExpression self,
      String ident){}

  /**
   * visits an instanceof expression
   */
  public void visitInstanceofExpression(JInstanceofExpression self,
				 JExpression expr, 
				 CType dest)
    {
	if( expr != null )
	    expr.accept(this);
    }

  /**
   * visits an equality expression
   */
     public void visitEqualityExpression(JEqualityExpression self,
			       int oper,
			       JExpression left,
			       JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits a conditional expression
   */
  public void visitConditionalExpression(JConditionalExpression self,
				  JExpression cond,
				  JExpression left,
				  JExpression right)
    {
	if( cond != null ) cond.accept(this );
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits a compound expression
   */
  public void visitCompoundAssignmentExpression(JCompoundAssignmentExpression self,
					 int oper,
					 JExpression left,
					 JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits a field expression
   */
  public void visitFieldExpression(JClassFieldExpression self,
			    JExpression left, 
			    String ident)
    {
	if( left != null ) left.accept(this);
    }

  /**
   * visits a class expression
   */
  public void visitClassExpression(JClassExpression self,
      CType type){}

  /**
   * visits a cast expression
   */
  public void visitCastExpression(JCastExpression self,
			   JExpression expr, 
			   CType type)
    {
	if( expr != null ) expr.accept(this);
    }

  /**
   * visits a cast expression
   */
  public void visitUnaryPromoteExpression(JUnaryPromote self,
				   JExpression expr, 
				   CType type)
    {
	if( expr != null ) expr.accept(this);
    }

  /**
   * visits a compound assignment expression
   */
  public void visitBitwiseExpression(JBitwiseExpression self,
			      int oper,
			      JExpression left,
			      JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }

  /**
   * visits an assignment expression
   */
  public void visitAssignmentExpression(JAssignmentExpression self,
				 JExpression left, 
				 JExpression right)
    {
	if( left != null ) left.accept(this);
	if( right != null ) right.accept(this);
    }


  /**
   * visits an array length expression
   */
  public void visitArrayLengthExpression(JArrayLengthExpression self,
				  JExpression prefix)
    {
	if( prefix != null ) prefix.accept(this);
    }

  /**
   * visits an array length expression
   */
  public void visitArrayAccessExpression(JArrayAccessExpression self,
				  JExpression prefix, 
				  JExpression accessor)
    {
	if( prefix != null ) prefix.accept(this);
	if( accessor != null ) accessor.accept(this);
    }

  /**
   * visits an array length expression
   */
    public void visitComments(JavaStyleComment[] comments){}

  /**
   * visits an array length expression
   */
    public void visitComment(JavaStyleComment comment){}

  /**
   * visits an array length expression
   */
    public void visitJavadoc(JavadocComment comment){}

  // ----------------------------------------------------------------------
  // OTHERS
  // ----------------------------------------------------------------------

  /**
   * visits an array length expression
   */
  public void visitSwitchLabel(JSwitchLabel self,
			JExpression expr)
    {
	if( expr != null ) expr.accept(this);
    }

  /**
   * visits an array length expression
   */
  public void visitSwitchGroup(JSwitchGroup self,
			JSwitchLabel[] labels, 
			JStatement[] stmts)
    {
	if( stmts != null )
	    for( int i = 0; i < stmts.length; i++ )
		stmts[i].accept(this);
    }

  /**
   * visits an array length expression
   */
  public void visitCatchClause(JCatchClause self,
			JFormalParameter exception, 
			JBlock body)
    {
	if( body != null )
	    body.accept(this);
    }

  /**
   * visits an array length expression
   */
  public void visitFormalParameters(JFormalParameter self,
			     boolean isFinal, 
			     CType type, 
			     String ident)
    {}
	

  /**
   * visits an array length expression
   */
  public void visitExplicitConstructorInvocation(JExplicitConstructorInvocation self,
					  JExpression prefix,
					  String ident,
					  JExpression[] params)
    {
	if (prefix != null ) prefix.accept(this);
	if( params != null )
	    for( int i = 0; i < params.length; i++ )
		params[i].accept(this);
    }

  /**
   * visits an array initializer expression
   */
  public void visitArrayInitializer(JArrayInitializer self,
			     JExpression[] elems)
    {
	if( elems != null )
	    for( int i = 0 ; i< elems.length; i++ )
		elems[i].accept(this);
    }

  /**
   * visits an array dimension
   */
  public void visitArrayDimsAndInit(JArrayDimsAndInits self,
			     JExpression[] dims, 
			     JArrayInitializer init)
    {
	if( dims != null )
	    for( int i = 0; i < dims.length; i++ )
	    {
		if( dims[i] != null ) dims[i].accept(this);
	    }
	if( init != null )
	    init.accept(this );
    }

  /**
   * visits a boolean literal
   */
    public void visitBooleanLiteral(boolean value){}

  /**
   * visits a byte literal
   */
    public void visitByteLiteral(byte value){}

  /**
   * visits a character literal
   */
    public void visitCharLiteral(char value){}

  /**
   * visits a double literal
   */
    public void visitDoubleLiteral(double value){}

  /**
   * visits a float literal
   */
    public void visitFloatLiteral(float value){}

  /**
   * visits a int literal
   */
    public void visitIntLiteral(int value){}

  /**
   * visits a long literal
   */
    public void visitLongLiteral(long value){}

  /**
   * visits a short literal
   */
    public void visitShortLiteral(short value){}

  /**
   * visits a string literal
   */
    public void visitStringLiteral(String value){}

  /**
   * visits a null literal
   */
    public void visitNullLiteral(){}

  /**
   * visits a package name declaration
   */
    public void visitPackageName(String name){}

  /**
   * visits a package import declaration
   */
    public void visitPackageImport(String name){}

  /**
   * visits a class import declaration
   */
   public  void visitClassImport(String name){}
}
