/* Jex - A tool for analyzing exception flow in Java programs
 *
 *  Copyright (C) 2001 Martin P. Robillard (mrobilla@cs.ubc.ca)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.ubc.cs.jex.util;

/** This class is use to represent a Java
  * package installed on a system, visible
  * from the current classpath.
  *
  * @author $Author: mrobilla $
  * @version $Revision: 1.2 $ $Date: 2002/03/08 21:56:23 $
  */ 

import java.util.StringTokenizer;
import java.io.File;
import java.io.IOException;

public class Package
{
    private String aName;
    private String aSource;
    private String[] aClasses;

    /** Loads a new package and create new object.
      * @param pName Fully qualified name of the package to load,
      * e.g., java.lang
      * @exception PackageNotFoundException If the package cannot be loaded from
      * the system
      */
    public Package( String pName ) throws PackageNotFoundException
    {
	aName = pName;
	
	StringTokenizer lClassPath = new StringTokenizer( System.getProperty("java.class.path"), File.pathSeparator );
	while( lClassPath.hasMoreTokens() )
	{
	    try
	    {
		String lCurrentPath = lClassPath.nextToken();
		PackageAccessor lAccessor = PackageAccessor.createPackageAccessor( new File( lCurrentPath ));
		aClasses = lAccessor.getClasses( aName );
		if( aClasses != null )
		{
		    aSource = lCurrentPath;
		    break;
		}
	    }
	    catch( IOException e )
	    {
		// An IOException here means fatal error reading a package;
		// Skip entry altogether.
		aClasses = null;
	    }
	}

	if( aClasses == null )
	    throw new PackageNotFoundException( "Could not load package " + aName );
    }

    public String toString()
    {
	return getName();
    }
    
    /** Returns the name of the package.
      * @return The fully qualified name of the package, e.g., "java.lang".
      */
    public String getName()
    { 
	return aName;
    }

    public boolean equals( Object pPackage )
    {
	boolean lReturn = false;
	if( pPackage instanceof Package )
	    if( ((Package)pPackage).getName().equals( this.getName() ))
		lReturn = true;
	return lReturn;
    }

    /** Returns the name of the source file where the package can be found.
      * @return The absolute file name where this package is located.
      */
    public String getSourceFileName()
    {
	return aSource;
    }
    
    /** Returns a list of all the classes in the package.
      * This list is determined by all the files with the .class
      * extension that are located in the directory corresponding
      * to the package.
      */
    public String[] getClasses()
    {
	return aClasses;
    }
}
